##################################################################################
#
#   Kaivo Public Software License
#
#   Copyright (c) 2001, Kaivo, Inc.
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#       o Redistributions of source code must retain the above copyright notice,
#         this list of conditions and the following disclaimer.
#
#       o Redistributions in binary form must reproduce the above copyright
#         notice, this list of conditions and the following disclaimer
#         in the documentation and/or other materials provided with the
#         distribution.
#
#       o Neither Kaivo nor the names of its contributors may be used to endorse
#         or promote products derived from this software without specific prior
#         written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#   ARE DISCLAIMED. IN NO EVENT SHALL KAIVO OR ITS CONTRIBUTORS BE LIABLE FOR 
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
##################################################################################

__doc__ = """Document Library Document Storage"""
__version__ = '0.5a'

# Document Store by Casey Duncan (cduncan@kaivo.com)
# Kaivo, Inc. (http://www.kaivo.com)

# This initial implementation uses a BTreeFolder to store the documents
# in the ZODB. Later implementations will abstract this to allow the
# option of storing the documents in a separate file system.

from string import strip, split, join, find
from StringIO import StringIO
import Globals
from Globals import DTMLFile
from App.ImageFile import ImageFile
from Products.BTreeFolder.BTreeFolder import BTreeFolder
from CatalogPlus import CatalogPlus
from DateTime.DateTime import DateTime
from ZPublisher.Converters import type_converters
import DLGlobals
import Document

manage_addLibraryDocumentStoreForm = DTMLFile('dtml/AddDocumentStoreForm', globals())

def manage_addLibraryDocumentStore(self, id, title='', vocab_id = None, 
                                   REQUEST=None):
    """Add a document store instance to a library"""
    ob = DocumentStore(id, title, vocab_id, self)
    self._setObject(id, ob)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_main')


class DocumentStore(BTreeFolder):
    """Document library document storage class"""

    meta_type="Document Store"
    all_meta_types=({'name': 'Document File', 'action': 'manage_addDocumentFileForm'},)

    manage_options = (
        {'label': 'Contents', 'action': 'manage_main'},
        {'label': 'Document Properties', 'action': 'manage_documentProperties'},
        {'label': 'Catalog Indexes', 'action': 'manage_catalogIndexes'},
        {'label': 'Catalog Metadata', 'action': 'manage_catalogColumns'},
        {'label': 'Catalog Maintenance', 'action': 'manage_catalogMaintenance'},
        {'label': 'Security', 'action': 'manage_access'},
        {'label': 'Undo', 'action': 'manage_UndoForm'},
        {'label': 'Owner', 'action': 'manage_owner'},
        {'label': 'Find', 'action': 'manage_findForm'},
    )

    __ac_permissions__ = BTreeFolder.__ac_permissions__ + (
        ('Add Document Library Files', ('manage_addDocumentFileForm', 'manage_addDocumentFile',
                                        'addDocumentFile')),
        ('Add Document Library Files Without Review', ()),
        ('Manage properties', ('manage_documentProperties',)),
        ('Delete objects', ('deleteDocumentFile',)),
        ('Search ZCatalog', ('query', 'getpath', 'getCatalogIndexes', 'getCatalogColumns')),
        ('Manage ZCatalog Entries', ('manage_reinitCatalog', 'manage_catalogColumns',
                                     'manage_delCatalogColumns', 'manage_addCatalogColumn',
                                     'manage_catalogIndexes', 'manage_delCatalogIndexes',
                                     'manage_addCatalogIndex','manage_catalogMaintenence')),        
    )

    _documentProperties = (
        { 'id': 'title', 'type': 'string', 'required': 1, 'mode': 'w' },
        { 'id': 'content_type', 'type': 'string', 'hidden': 1, 'mode': 'w' },
        { 'id': 'filename', 'type': 'string', 'hidden': 1, 'default': '', 'mode': 'w'},
        { 'id': 'type', 'type': 'selection', 'select_variable': 'document_types', 'default':'', 'required': 1 },
        { 'id': 'description', 'type': 'text', 'default': '', 'required': 1 },
        { 'id': 'topics', 'type': 'lines', 'default': [], 'required': 1, 'mode': 'w' },
        { 'id': 'creator', 'type': 'string', 'default': '' },
        { 'id': 'date', 'type': 'date', 
          'default': DLGlobals.document_no_creation_date, 'mode': 'w' },
        { 'id': 'review_date', 'type': 'date', 
          'default': DLGlobals.document_unreviewed_date, 'mode': 'w' },
    )

    _defaultCatalogColumns = ( 'id', 'title', 'identifier', 'date', 'description' )
    _defaultCatalogIndexes = (
        ('id', 'FieldIndex'),
        ('title', 'TextIndex'),
        ('type', 'FieldIndex'),
        ('description', 'TextIndex'),
        ('topics', 'KeywordIndex'),
        ('super_topics', 'KeywordIndex'),
        ('creator', 'TextIndex'),
        ('date', 'FieldIndex'),
        ('review_date', 'FieldIndex'),
        ('all_searchable_text', 'TextIndex'),
        ('owner_id', 'FieldIndex'),
    )

    def __init__(self, id, title='', vocab_id=None, container=None):
        self.id = id
        self.title = title

        if vocab_id is not None and container is None:
            raise AttributeError, 'You cannot specify a vocab_id without' \
                                  'also specifying a container.'
        if container is not None:
            self=self.__of__(container)
        
        # Setup the document store schema and indexes
        self.Catalog = CatalogPlus(vocab_id)
        
        for name in self._defaultCatalogColumns:
            self.Catalog.addColumn(name)

        for name, type in self._defaultCatalogIndexes:
            self.Catalog.addIndex(name, type)
        
        try: # Be flexible with the BTreeFolder implementation
            BTreeFolder.__init__(self)
        except:
            BTreeFolder.__init__(self, id)

    DocumentStore_icon = ImageFile('www/DocumentStore_icon.gif', globals())
    def icon(self):
        """Managment icon"""
        return join(self.getPhysicalPath()[1:],'/') + '/DocumentStore_icon'

    #
    # Document property methods
    #

    def defaultPropertyValue(self, propertyId, d=None):
        """Return the default value for a document property"""
        for p in self._documentProperties:
            if p['id'] == propertyId:
                try:
                    return p['default']
                except KeyError:
                    return d
        return d        

    def documentPropertyMap(self):
        """Return mapping of document property schema"""
        return self._documentProperties

    def documentPropertyDict(self):
        """Return a dictionary of property info keyed on id"""
        propdict = {}

        for prop in self._documentProperties:
            propdict[prop['id']] = prop

        return propdict

    def documentPropertyErrors(self, REQUEST, error_template='<span class="form-error">%s</span>'):
        """Validates document properties passed on the REQUEST object
           returns an error string if errors are found and an empty string otherwise"""
        errors = {}
        for p in self._documentProperties:
            v = REQUEST.get(p['id'], None)

            if p.has_key('required') and p['required'] and not v:
                errors[p['id']] = error_template % 'Required Field'
            elif p['type'] == 'date' and v:
                try:
                    date = DateTime(v)
                    REQUEST.set(p['id'], v)
                except:
                    errors[p['id']] = error_template % 'Invalid Date'

        if errors:
            REQUEST.set('property_errors', errors)
            return error_template % "Errors were found on the form." \
                                    "please correct the errors and resubmit."
        else:
            return ''

    manage_documentProperties = DTMLFile('dtml/EditDocumentProperties', globals())

    def manage_editDocumentProperties(self, REQUEST, RESPONSE=None):
        """Modify the document properties' default values"""
        
        for prop in self._documentProperties:
            propvalue = REQUEST.get(prop['id'], '')
            proptype = prop.get('type', 'string')

            if type_converters.has_key(proptype):
                propvalue = type_converters[proptype](propvalue)

            prop['default'] = propvalue
    
            required_flag = prop['id'] + '_required_'
            
            if REQUEST.has_key(required_flag) and REQUEST[required_flag]:
                prop['required'] = 1
            else:
                if prop.has_key('required'): del prop['required']

        self._p_changed = 1

        if REQUEST and RESPONSE:
            RESPONSE.redirect(REQUEST.URL1 + ('/manage_documentProperties?'
                                              'manage_tabs_message=Properties+Updated'))

    def manage_delDocumentProperties(self, ids=None, REQUEST=None, RESPONSE=None):
        """Delete one or more document properties specified by 'ids'."""
        if not ids:
            raise 'BadRequest', "No properties were specified"

        propdict=self.documentPropertyDict()
        nd=self._reserved_names
        for id in ids:
            if not propdict.has_key(id):
                raise 'BadRequest', (
                      'The property <em>%s</em> does not exist' % id)
            if (not 'd' in propdict[id].get('mode', 'wd')) or (id in nd):
                raise 'BadRequest', (
                      'The property <em>%s</em> cannot be deleted.' % id)
            del propdict[id]

        self._documentProperties = tuple(map(lambda i: i[1], propdict.items()))
   
        if REQUEST and RESPONSE:
            RESPONSE.redirect(REQUEST.URL1 + ('/manage_documentProperties'
                                              '?manage_tabs_message=Properties+Deleted'))

    def manage_addDocumentProperty(self, id, default_value, type, required=0,
                                  REQUEST=None, RESPONSE=None):
        """Add a document property"""
        # Raise an error if an value is wrapped.
        if hasattr(default_value, 'aq_base'):
            raise ValueError, 'Invalid property value: wrapped object'

        # Check that the property id is valid
        docpropnames = map(lambda p: p['id'], self._documentProperties)

        if not id or id[:1]=='_' or (id[:3]=='aq_') \
           or (' ' in id) or hasattr(self.aq_base, id) or (id in docpropnames):
            raise 'Bad Request', 'Invalid or duplicate property id'

        # Convert the value to the proper type
        if type_converters.has_key(type):
            default_value = type_converters[type](default_value)

        prop = {}
        prop['id'] = id
        prop['type'] = type

        if not type in ('selection', 'multiple selection'):
            prop['default'] = default_value
        else:
            prop['select_variable'] = default_value
            prop['default'] = ''

        if required: prop['required'] = 1
        self._documentProperties = self._documentProperties + (prop,)

        if REQUEST and RESPONSE:
            RESPONSE.redirect('manage_documentProperties?manage_tabs_message=Property+Added')

    #
    # Document management methods
    #

    manage_addDocumentFileForm = Document.manage_addDocumentFileForm
    manage_addDocumentFile = Document.manage_addDocumentFile

    def addDocumentFile(self, REQUEST):
        """Adds a new document and sets its properties from a form based submission"""

        self=self.this()
        doc = Document.Document(title=REQUEST.title, container=self)
        self._setObject(doc.getId(), doc)
        doc = self._getOb(doc.getId())

        # See if the user can skip the document review
        if REQUEST.AUTHENTICATED_USER.has_permission('Add Document Library Files Without Review',
                                                     self):
            REQUEST.set('review_date', DateTime())
        else:
            REQUEST.set('review_date', DLGlobals.document_unreviewed_date)

        # Set the properties, but don't index yet
        doc.manage_changeProperties(REQUEST, reindex=0)
        # See if a file or url was specified
        if not REQUEST.file.filename and REQUEST.has_key('url') and REQUEST.url and REQUEST.url != 'http://':
            doc.manage_upload(file=REQUEST.url, remote=1)
        else:
            doc.manage_upload(REQUEST.file)

        if not Document.check_content_type(self, doc.content_type):
            self.manage_delObjects(doc.getId())
            raise "IllegalFileType", \
                  "<span>That file type cannot be uploaded into the library.</span>"

        return doc

    def deleteDocumentFile(self, REQUEST=None, RESPONSE=None):
        """Deletes a document from the document store"""
        self.manage_delObjects(REQUEST.id)

        if REQUEST and RESPONSE and REQUEST.has_key('last_url'):
            RESPONSE.redirect(REQUEST.last_url + '?' + REQUEST.get('last_query',''))

    #
    #   Catalog query support methods
    #

    def query(self, REQUEST=None, **kw):
        """Query the document store for matching documents"""
        return apply(self.Catalog.searchResults, (REQUEST,), kw)

    def getpath(self, rid):
        """
        Return the path to a cataloged object given a 'data_record_id_'
        Used by the Catalog brains
        """
        return self.Catalog.paths[rid]

    #
    # Catalog management methods
    #

    manage_catalogColumns = DTMLFile('dtml/catalogColumns', globals())

    def getCatalogColumns(self):
        """Returns the metadata columns in the catalog"""
        return self.Catalog.schema.keys()

    def manage_delCatalogColumns(self, names, REQUEST=None, RESPONSE=None):
        """Delete Catalog metadata column"""
        if type(names) == type(''):
            names = [names]

        for name in names:
            self.Catalog.delColumn(name)

        if REQUEST and RESPONSE:
            RESPONSE.redirect('manage_catalogColumns?manage_tabs_message=Catalog+Metadata+Deleted')

    def manage_addCatalogColumn(self, name, REQUEST=None, RESPONSE=None):
        """Add Catalog metdata column"""
        self.Catalog.addColumn(name)

        if REQUEST and RESPONSE:
            RESPONSE.redirect('manage_catalogColumns?manage_tabs_message=Catalog+Metadata+Column+Added')

    manage_catalogIndexes = DTMLFile('dtml/catalogIndexes', globals())

    def getCatalogIndexes(self):
        """Returns the indexes in the catalog"""
        return self.Catalog.indexes.values()

    def manage_delCatalogIndexes(self, names, REQUEST=None, RESPONSE=None):
        """Delete Catalog index"""
        if type(names) == type(''):
            names = [names]

        for name in names:
            self.Catalog.delIndex(name)

        if REQUEST and RESPONSE:
            RESPONSE.redirect('manage_catalogIndexes?manage_tabs_message=Catalog+Index+Deleted')

    def manage_addCatalogIndex(self, name, type, REQUEST=None, RESPONSE=None):
        """Add Catalog index"""
        self.Catalog.addIndex(name, type)

        if REQUEST and RESPONSE:
            RESPONSE.redirect('manage_catalogIndexes?manage_tabs_message=Catalog+Index+Added')
        
    manage_catalogMaintenance = DTMLFile('dtml/catalogMaintenance', globals())

    def manage_reinitCatalog(self, REQUEST=None, RESPONSE=None):
        """Reinitialize the document store catalog"""

        self.Catalog.clear()
       
        for doc in self.objectValues():
            doc.setIcons()
            if doc.is_accepted(): doc.showTopics()
            doc.index_object()

        if REQUEST and RESPONSE:
            RESPONSE.redirect('manage_catalogMaintenance?manage_tabs_message=Catalog+Reinitialized')

Globals.InitializeClass(DocumentStore)
